import os
import time
import timeit
from .global_config import get_timer

def timer(func):
    """
    Décorateur pour mesurer le temps d'exécution d'une fonction et enregistrer ce temps dans un logger si activé.
    :param func: La fonction à décorer pour mesurer et enregistrer son temps d'exécution.
    :raises ValueError: Si aucun logger n'est défini dans les arguments de la fonction appelée.
    :return: La fonction décorée qui mesure le temps d'exécution.
    """
    def wrapper(*args, **kwargs):
        if get_timer():
            logger = kwargs.get('logger', None)
            if logger is None and args and hasattr(args[0], 'logger'):
                logger = args[0].logger
            if logger is None: raise ValueError("Pas de logger défini.")
            start_time = timeit.default_timer()
            result = func(*args, **kwargs)
            elapsed_time = timeit.default_timer() - start_time
            if logger is not None:
                logger.info(f"Temps d'exécution de {func.__name__}: {elapsed_time:.4f} secondes.")
            return result
        else:
            return func(*args, **kwargs)
    return wrapper

def set_timezone(tz = 'Europe/Paris'):
    os.environ['TZ'] = tz
    time.tzset()


import os
import time

def wait_for_file(path, filename, retry=20):
    """
    Attends qu'un fichier spécifique soit disponible dans un répertoire donné, avec un nombre limité de tentatives.
    :param path: Le chemin du répertoire où chercher le fichier.
    :param filename: Le nom du fichier à attendre dans le répertoire spécifié.
    :param retry: Le nombre maximal de tentatives pour vérifier la présence du fichier (par défaut 20).
    :raises ValueError: Si le chemin spécifié dans `path` n'existe pas.
    :return: Retourne True si le fichier est trouvé dans le nombre de tentatives, sinon False.
    """
    i=1
    res = False
    if not os.path.exists(path):
        raise "Error : path does not exist"
    while not os.path.isfile(os.path.join(path, filename)) and i <= retry:
        time.sleep(1)
        i += 1
    if i <= retry:
        res = True
    return res



from .timer import timer
from .logger import Logger
import polars as pl
from polars.testing import assert_frame_equal

class dataComparator:
    def __init__(self, dictionnary: dict, logger: Logger):
        """
        Initialise un dataComparator avec les informations de connexions aux bases de données
        :param dictionnary: le dictionnaire qui contient les informations du pipeline
            - 'db_source_1': la base de données source 1
            - 'query_source_1': la requête à envoyer à la source 1
            - 'db_source_2': la base de données source 2
            - 'query_source_2': la requête à envoyer à la source 2
            - 'batch_size': la taille des lots pour le traitement en batch
        :param logger: le logger pour gérer la journalisation des évènements du pipeline
        """
        self.logger = logger
        self.__db_source_1 = dictionnary.get('db_source_1')
        self.__query_source_1 = dictionnary.get('query_source_1')
        self.__db_source_2 = dictionnary.get('db_source_2')
        self.__query_source_2 = dictionnary.get('query_source_2')
        self.__batch_size = dictionnary.get('batch_size', 10_000)

    def _fetch_data(self, db, query):
        """
        Exécute la requête donnée sur la base spécifiée et retourne un DataFrame Polars
        :param db: la base sur laquelle envoyer la requête
        :param query: la requête à envoyer
        :return: le résultat de la requête sous forme de DataFrame
        """
        db.connect()
        data = list(db.sqlQuery(query))[0]
        return pl.DataFrame(data, orient='row', strict=False, infer_schema_length=10_000)

    @timer
    def compare(self):
        """
        Compare les résultats des deux requêtes SQL entre les deux bases de données.
        :return: Un dictionnaire contenant les différences si elles existent
        """
        try:
            data_1 = self._fetch_data(self.__db_source_1, self.__query_source_1)
            data_2 = self._fetch_data(self.__db_source_2, self.__query_source_2)
            assert_frame_equal(data_1, data_2)
        except Exception as e:
            self.logger.error(f"Erreur lors de la comparaison des données: {e}")
            raise



import os
import dotenv

class env:
    def __init__(self, logger, file=None):
        """
        Charge les variables d'environnement à partir d'un fichier `.env`.
        Si aucune fichier n'est spécifié et que `.env` n'est pas trouvé, on se contente des variables
        d'environnement déjà définis
        :param logger: Le Logger est un objet d'adsToolBox. Veuillez définir un Logger() avant de faire appel à cette méthode.
        :param file: Le chemin vers un fichier `.env` spécifique à charger. Si ce paramètre est omis,
            la classe cherche automatiquement un fichier `.env` à la racine du projet.
        """
        if file:
            if os.path.isfile(file):
                dotenv_file = dotenv.find_dotenv(file)
                dotenv.load_dotenv(dotenv_file)
                logger.debug("Fichier spécifié pour l'environnement trouvé")
            else:
                logger.warning("Fichier spécifié non trouvé")
        elif os.path.isfile('.env'):
            dotenv_file = dotenv.find_dotenv('.env')
            dotenv.load_dotenv(dotenv_file)
        else:
            logger.warning("Fichier .env non trouvé à la racine du projet")
            logger.warning("Veuillez créer un fichier .env à la racine du projet ou en spécifier un vous même")
        for cle, valeur in os.environ.items():
            setattr(self, cle, valeur)


from .logger import Logger
from .timer import timer, get_timer, set_timezone
from .global_config import set_timer
from .dbPgsql import dbPgsql
from .dbMssql import dbMssql
from .dbMysql import dbMysql
from .pipeline import pipeline
from .wait_file import wait_for_file
from .loadEnv import env
from .odoo import OdooConnector
from .dataComparator import dataComparator
from .mail import mail
from .file_handler import FileHandler
from .cdc import ChangeDataCapture

# Imports des méthodes et classes de la librairie
__all__ = [
    "Logger", "timer", "set_timer", "get_timer", "dbPgsql", "dbMssql", "pipeline", "OdooConnector",
    "wait_for_file", "env", "dataComparator", "dbMysql", "mail", "FileHandler", "ChangeDataCapture",
    "set_timezone"
]


import pytds
import pymssql
import timeit
import polars as pl
from .timer import timer, get_timer
from .dataFactory import data_factory

class dbMssql(data_factory):
    def __init__(self, dictionnary: dict, logger, batch_size=10_000):
        """
        instancie la classe dbMssql, qui hérite de dataFactory
        :param dictionnary: Un dictionnaire contenant tous les paramètres nécéssaires pour lancer une connexion sql server
        :param logger: un logger ads qui va gérer les logs des actions de la classe
        :param batch_size: la taille des batchs en lecture et écriture
        :param package: le package à utiliser ('pymssql' qui gère les connexion encryptées mais qui encode mal des
            caractères spéciaux en bulk ou alors 'pytds' qui est moins permissif et ne gère pas les connexion encryptées
            mais n'encode pas mal les caractères spéciaux
        """
        self.connection = None
        self.logger = logger
        self.__database = dictionnary.get('database')
        self.__user = dictionnary.get('user')
        self.__password = dictionnary.get('password')
        self.__port = dictionnary.get('port')
        self.__host = dictionnary.get('host')
        self.__charset = dictionnary.get('charset', 'UTF-8')
        self.batch_size = batch_size
        self.__package = dictionnary.get('package', 'pymssql')


    @timer
    def connect(self):
        """
        lance la connexion avec les identifiants passés à l'initialisation de la classe
        toutes les méthodes de la classe nécéssitent une connexion active
        :return: la connexion
        """
        if self.logger is not None: self.logger.info("Tentative de connexion avec la base de données.")
        try:
            if self.__package == 'pytds':
                self.connection = pytds.connect(
                    dsn=self.__host,
                    port=self.__port,
                    user=self.__user,
                    password=self.__password,
                    database=self.__database,
                    autocommit=False,
                )
                self.sqlExec("SET DATEFORMAT ymd;")
            elif self.__package == 'pymssql':
                server = f"{self.__host}:{self.__port}" if self.__port else self.__host
                self.connection = pymssql.connect(
                    server=server,
                    user=self.__user,
                    password=self.__password,
                    database=self.__database,
                    charset=self.__charset
                )
            else:
                raise ValueError(f"Le package {self.__package} n'est pas supporté.")
            if self.logger is not None: self.logger.info(f"Connexion établie avec la base de données via {self.__package}.")
            return self.connection
        except Exception as e:
            if self.logger is not None: self.logger.error(f"Échec de la connexion à la base de données: {e}")
            raise

    def sqlQuery(self, query: str, return_columns: bool = False):
        """
        lit la base de données avec la requête query
        :param return_columns: booléen qui indique si l'on veut récupérer les colonnes de la tables
        :param query: la requête
        :return: les données lues avec yield
        """
        self.logger.debug(f"Exécution de la requête de lecture : {query}")
        try:
            timer_start = timeit.default_timer()
            cpt_rows = 0
            with self.connection.cursor() as cursor:
                cursor.execute(query)
                if return_columns:
                    cols = [desc[0] for desc in cursor.description]
                    yield cols
                self.logger.info("Requête exécutée avec succès, début de la lecture des résultats.")
                while True:
                    rows = cursor.fetchmany(self.batch_size)
                    if not rows:
                        break
                    yield rows
                    cpt_rows+=len(rows)
                    self.logger.info(f"{cpt_rows} ligne(s) lue(s).")
            if get_timer():
                elapsed_time = timeit.default_timer() - timer_start
                self.logger.info(f"Temps d'exécution de sqlQuery: {elapsed_time:.4f} secondes")
        except Exception as e:
            self.logger.error(f"Échec de la lecture des données: {e}")
            raise

    @timer
    def sqlExec(self, query):
        """
        execute une requête sur la base de données, un create ou delete table par exemple
        :param query: la requête
        """
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query)
                self.connection.commit()
                self.logger.info(f"Requête exécutée avec succès.")
        except Exception as e:
            self.logger.error(f"Échec de l'exécution de la requête: {e}")
            raise

    @timer
    def sqlScalaire(self, query):
        """
        execute une requête et retourne le premier résultat
        :param query: la requête
        :return: le résultat de la requête
        """
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query)
                result = cursor.fetchone()
                self.logger.info(f"Requête scalaire exécutée avec succès.")
                return result[0] if result else None
        except Exception as e:
            self.logger.error(f"Échec de l'exécution de la requête: {e}")
            raise

    @staticmethod
    def __escape_value(value):
        if isinstance(value, str):
            return value.replace('"', '""').replace("'", "''")
        return value

    @timer
    def insert(self, schema: str, table: str, cols: [], row: []):
        """
        Insère des données dans la base de données
        :param schema: Nom du schéma dans lequel est la table
        :param table: Nom de la table dans laquelle insérer
        :param cols: liste des colonnes dans lesquelles insérer
        :param row: liste des valeurs à insérer
        :return: le résultat de l'opération, l'erreur et la la ligne concernée en cas d'erreur
        """
        placeholders = ", ".join(["%s"] * len(cols))
        table = f"{schema}.{table}" if schema else table
        query = f"INSERT INTO {table} ({', '.join(cols)}) VALUES ({placeholders})"
        row = [self.__escape_value(value) for value in row]
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query, row)
                self.connection.commit()
                self.logger.info(f"{len(row)} valeur(s) insérée(s) avec succès dans la table {table}")
                return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Échec de l'insertion des données: {e}")
            return "ERROR", str(e), row

    @timer
    def insertMany(self, schema: str, table: str, cols: [], rows: [[]]):
        """
        Insère des données par batch dans une table avec gestion des erreurs

        :param schema: Nom du schéma dans lequel se trouve la table
        :param table: Nom de la table dans laquelle insérer
        :param cols: Liste des colonnes dans lesquelles insérer
        :param rows: Liste des lignes à insérer
        :return: Le résultat de l'opération, l'erreur et la le batch concerné en cas d'erreur
        """
        failed_batches = []
        errors = []
        total_inserted = 0
        placeholders = ', '.join(["%s"] * len(cols))
        table = f"{schema}.{table}" if schema else table
        query = f"INSERT INTO {table} ({', '.join(cols)}) VALUES ({placeholders})"
        rows = [[self.__escape_value(value) for value in row] for row in rows]
        try:
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, len(rows), self.batch_size), start=1):
                    batch = rows[start: start + self.batch_size]
                    try:
                        cursor.executemany(query, batch)
                        self.connection.commit()
                        total_inserted += len(batch)
                        self.logger.info(
                            f"Batch {batch_index}: {len(batch)} ligne(s) insérée(s) avec succès dans la table {table}. "
                            f"Total inséré: {total_inserted}/{len(rows)} ligne(s)."
                        )
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'insertion du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
                return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'insertion des données: {e}")
            return "ERROR", str(e), rows

    def __get_df(self, rows, cols):
        if isinstance(rows, list):
            df = pl.DataFrame(rows, schema=cols, orient='row', infer_schema_length=len(rows))
        elif isinstance(rows, pl.DataFrame):
            df = rows
        else:
            raise ValueError("Les données doivent être une liste de tuples ou un DataFrame polars")
        return df

    def __insertBulk_pymssql(self, schema: str, table: str, cols: [], rows):
        import json
        import datetime
        failed_batches = []
        errors = []
        total_inserted = 0
        table_full = f"{schema}.{table}" if schema else table
        df = self.__get_df(rows, cols)
        type_mapping= {}
        for col in cols:
            non_null_values = df[col].drop_nulls()
            sample_value = non_null_values[0] if len(non_null_values) > 0 else 0
            type_mapping[col] = (
                "DATETIME2" if isinstance(sample_value, datetime.date)
                else "TIME(3)" if isinstance(sample_value, datetime.time)
                else "NVARCHAR(MAX)"
            )
        col_defs = ", ".join([f"[{col}]" for col in cols])
        openjson_defs = ", ".join([f"[{col}] {type_mapping[col]} '$.{col}'" for col in cols])
        query = f"""
        INSERT INTO {table_full} ({col_defs})
        SELECT {col_defs}
        FROM OPENJSON(%s)
        WITH ({openjson_defs});
        """
        try:
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, df.height, self.batch_size), start=1):
                    batch = df.slice(start, self.batch_size)
                    try:
                        as_json = [row for row in batch.iter_rows(named=True)]
                        json_data = json.dumps(as_json, default=str)
                        cursor.execute(query, (json_data,))
                        self.connection.commit()
                        total_inserted += len(batch)
                        self.logger.info(
                            f"Batch {batch_index}: {len(batch)} ligne(s) insérée(s) avec succès dans la table {table_full}. "
                            f"Total inséré: {total_inserted}/{len(rows)} ligne(s)."
                        )
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'insertion du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
            return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'insertion des données: {e}")
            return "ERROR", str(e), rows

    def __insertBulk_pytds(self, schema: str, table: str, cols: [], rows):
        failed_batches = []
        errors = []
        total_inserted = 0
        table_full = f"{schema}.{table}" if schema else table
        try:
            df = self.__get_df(rows, cols)
            n_rows = df.shape[0]
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, n_rows, self.batch_size), start=1):
                    batch = df.slice(start, self.batch_size)
                    data = [tuple(str(value) if value else None for value in row) for row in
                            batch.iter_rows(named=False)]
                    try:
                        table, schema = (s[1:-1] if '[' in s else s for s in (table, schema))
                        cursor.copy_to(
                            data=data,
                            schema=schema,
                            table_or_view=table,
                            columns=cols
                        )
                        self.connection.commit()
                        total_inserted += batch.shape[0]
                        self.logger.info(
                            f"Batch {batch_index}: {batch.shape[0]} ligne(s) insérée(s) avec succès dans la table {table_full}. "
                            f"Total inséré: {total_inserted}/{n_rows} ligne(s).")
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'insertion du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
            return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'insertion des données: {e}")
            return "ERROR", str(e), rows

    @timer
    def insertBulk(self, schema: str, table: str, cols: [], rows):
        """
        Insère des données dans une table en bulk, en lots

        :param schema: Nom du schéma
        :param table: Nom de la table dans laquelle insérer
        :param cols: Liste des colonnes dans lesquelles insérer
        :param rows: Liste des lignes à insérer
        :return: Le résultat de l'opération, l'erreur et la le batch concerné en cas d'erreur
        """
        if self.__package == "pymssql":
            result, errors, failed_batches = self.__insertBulk_pymssql(schema, table, cols, rows)
        elif self.__package == "pytds":
            result, errors, failed_batches = self.__insertBulk_pytds(schema, table, cols, rows)
        else:
            msg = f"Le package {self.__package} n'est pas supporté. Tentez avec 'pymssql' ou 'pytds'"
            self.logger.error(msg)
            raise ValueError(msg)
        return result, errors, failed_batches

    @timer
    def upsert(self, schema :str, table: str, cols: [], row: [], conflict_cols: []):
        """
        Réalise une opération upsert sur la base
        :param schema: Le schéma de la table
        :param table: Le nom de la table
        :param cols: Liste des colonnes dans lequelles insérer
        :param row: Liste des valeurs à insérer
        :param conflict_cols: Colonnes à utiliser pour détecter les conflits
        :return: Le résultat de l'opération, l'erreur et la ligne en cas d'erreur
        """
        table = f"{schema}.{table}" if schema else table
        target_cols = ', '.join(cols)
        source_cols = ', '.join([f"source.{col}" for col in cols])
        update_clause = ', '.join([f"target.{col} = source.{col}" for col in cols if col not in conflict_cols])
        conflict_cdt = ' AND '.join([f"target.{col} = source.{col}" for col in conflict_cols])
        query = f"""
        MERGE INTO {table} AS target
        USING (VALUES ({', '.join(['%s'] * len(cols))})) AS source ({', '.join(cols)})
        ON {conflict_cdt}
        WHEN MATCHED THEN
            UPDATE SET {update_clause}
        WHEN NOT MATCHED THEN
            INSERT ({target_cols})
            VALUES ({source_cols});"""
        row = [self.__escape_value(value) for value in row]
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query, row)
                self.connection.commit()
                self.logger.info(f"{len(row)} valeur(s) insérée(s) ou mise(s) à jour dans la table {table}.")
                return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur lors de l'upsert dans la table {table}.")
            return "ERROR", str(e), row

    @timer
    def upsertMany(self, schema: str, table: str, cols: [], rows: [[]], conflict_cols: []):
        """
        Réalise une opération upsert par batch sur la base
        :param schema: Le schéma de la table
        :param table: Le nom de la table
        :param cols: Liste des colonnes dans lequelles insérer
        :param rows: Liste des lignes à insérer
        :param conflict_cols: Colonnes à utiliser pour détecter les conflits
        :return: Le résultat de l'opération, l'erreur et le batch en cas d'erreur
        """
        failed_batches = []
        errors = []
        total_inserted = 0
        table = f"{schema}.{table}" if schema else table
        target_cols = ', '.join(cols)
        source_cols = ', '.join([f"source.{col}" for col in cols])
        update_clause = ', '.join([f"target.{col} = source.{col}" for col in cols if cols not in conflict_cols])
        conflict_cdt = ' AND '.join([f"target.{col} = source.{col}" for col in conflict_cols])
        query = f"""
        MERGE INTO {table} AS target
        USING (VALUES ({', '.join(['%s'] * len(cols))})) AS source ({', '.join(cols)})
        ON {conflict_cdt}
        WHEN MATCHED THEN
            UPDATE SET {update_clause}
        WHEN NOT MATCHED THEN
            INSERT ({target_cols})
            VALUES ({source_cols});"""
        rows = [[self.__escape_value(value) for value in row] for row in rows]
        try:
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, len(rows), self.batch_size), start=1):
                    batch = rows[start: start + self.batch_size]
                    try:
                        cursor.executemany(query, batch)
                        self.connection.commit()
                        total_inserted += len(batch)
                        self.logger.info(
                            f"Batch {batch_index}: {len(batch)} ligne(s) insérée(s) ou mise(s) à jour dans la table {table}. "
                            f"Total inséré ou mis à jour: {total_inserted}/{len(rows)} ligne(s)."
                        )
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'upsert du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
                return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'upsert des données: {e}")
            return "ERROR", str(e), rows

    def __create_temp_table(self, schema: str, table: str, temp_table: str, cols: []):
        query = f"""
        SELECT COLUMN_NAME, DATA_TYPE, CHARACTER_MAXIMUM_LENGTH
        FROM INFORMATION_SCHEMA.COLUMNS
        WHERE TABLE_SCHEMA = '{schema}' AND TABLE_NAME = '{table}'
        """
        self.logger.info(f"Récupération des colonnes depuis {schema}.{table}")
        column_definitions = list(self.sqlQuery(query))[0]
        col_def = []
        for col in cols:
            col_info = next((c for c in column_definitions if c[0] == col), None)
            if col_info:
                col_name, data_type, char_length = col_info
                if char_length:
                    col_def.append(f"{col_name} {data_type.upper()}({char_length})")
                else:
                    col_def.append(f"{col_name} {data_type.upper()}")
            else:
                raise ValueError(f"Column {col} not found in {schema}.{table}")
        col_def_sql = ', '.join(col_def)
        self.sqlExec(f"DROP TABLE IF EXISTS {schema}.{temp_table}")
        self.sqlExec(f"CREATE TABLE {schema}.{temp_table} ({col_def_sql})")

    @timer
    def upsertBulk(self, schema: str, table: str, cols: [], rows, conflict_cols):
        """
        Réalise une opération upsert en bulk par batch sur la base en lots et avec l'aide d'une table temporaire
        :param schema: Le schéma de la table
        :param table: Le nom de la table
        :param cols: Liste des colonnes dans lequelles insérer
        :param rows: Liste des lignes à insérer
        :param conflict_cols: Colonnes à utiliser pour détecter les conflits
        :return: Le résultat de l'opération, l'erreur et le batch en cas d'erreur
        """
        temp_table = f"{table}_temp"
        table_full = f"{schema}.{table}" if schema else table
        conflict_cdt = ' AND '.join([f"target.{col} = source.{col}" for col in conflict_cols])
        update_clause = ', '.join([f"target.{col} = source.{col}" for col in cols if col not in conflict_cols])
        target_cols = ', '.join(cols)
        source_cols = ', '.join([f"source.{col}" for col in cols])
        query = f"""
        MERGE INTO {table_full} AS target
        USING {temp_table} AS source
        ON {conflict_cdt}
        WHEN MATCHED THEN
            UPDATE SET {update_clause}
        WHEN NOT MATCHED THEN
            INSERT ({target_cols})
            VALUES ({source_cols});
        """
        try:
            self.__create_temp_table(schema, table, temp_table, cols)
            result, errors, failed_batches = self.insertBulk(schema, temp_table, cols, rows)
            if result == "ERROR":
                return result, errors, failed_batches
            self.sqlExec(query)
            self.logger.info(f"{len(rows)} lignes insérée(s) ou mise(s) à jour dans la tables {table}.")
            return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'UPSERT: {e}")
            return "ERROR", str(e), rows
        finally:
            self.sqlExec(f"DROP TABLE IF EXISTS {schema}.{temp_table}")
            self.logger.info(f"Table temporaire supprimée.")



_timer_enabled = False  # Variable globale pour déterminer si le timer est actif ou non

def set_timer(state: bool):
    """
    Active ou désactive le timer
    :param state: détermine l'état du timer
    """
    global _timer_enabled
    _timer_enabled = state


def get_timer() -> bool:
    """
    récupère l'état du timer
    :return: l'état du timer
    """
    return _timer_enabled



from abc import ABC, abstractmethod

class data_factory(ABC):
    """
    classe abstraite qui pose un modèle de connexion ads
    """
    @abstractmethod
    def connect(self):
        """
        lance la connexion avec les identifiants passés à l'initialisation de la classe
        :return: la connexion
        """
        pass

    @abstractmethod
    def insert(self,table,cols=[],rows=[]):
        """
        insère des données dans la base de données, nécessite une connexion active
        :param table: nom de la table dans laquelle insérer
        :param cols: liste des colonnes dans lesquelles insérer
        :param rows: liste des valeurs à insérer
        """
        pass

    @abstractmethod
    def insertBulk(self,table,cols=[],rows=[]):
        """
        similaire à insert classique, mais insère par batch de taille 'batch_size' définie
        :param table: nom de la table dans laquelle insérer
        :param cols: liste des colonnes dans lesquelles insérer
        :param rows: liste des lignes à insérer
        """
        pass

    @abstractmethod
    def sqlQuery(self,query):
        """
        lit la base de données avec la requête query
        :param query: la requête
        :return: les données lues
        """
        pass

    @abstractmethod
    def sqlExec(self, query: str, params=None):
        """
        execute une requête sur la base de données
        :param query: la requête
        :param params: d'éventuels paramètres à ajouter si la requête contient des variables
        """
        pass


import os
import timeit
import unicodedata
import smbclient
import chardet
from .timer import timer, get_timer

class FileHandler:
    def __init__(self, logger, smb_config: dict=None, batch_size: int=4_096):
        """
        Initialise le gestionnaire de fichiers avec le chemin du fichier
        """
        self.logger = logger
        self.smb_config = smb_config
        self.batch_size = batch_size
        if smb_config:
            self._setup_smb_connection()

    @timer
    def _setup_smb_connection(self):
        """
        Configure la connection SMB
        """
        try:
            smbclient.register_session(
                self.smb_config.get("server"),
                username=self.smb_config.get("username"),
                password=self.smb_config.get("password")
            )
            self.logger.info("Connexion SMB établie.")
        except Exception as e:
            self.logger.error(f"Échec de la connexion SMB : {e}")
            raise

    def read_file(self, file_path: str, mode: str='rb', encoding: str=None):
        """
        Lit un fichier local ou SMB en batchs
        """
        if self.smb_config and file_path.startswith("//"):
            return self._read_smb_file(file_path, mode, encoding)
        return self._read_local_file(file_path, mode, encoding)

    def _read_smb_file(self, file_path, mode, encoding):
        """
        Lit un fichier sur un partage SMB en batchs
        """
        timer_start = timeit.default_timer()
        try:
            with smbclient.open_file(file_path, mode=mode, encoding=encoding) as file:
                while chunk := file.read(self.batch_size):
                    yield chunk
                self.logger.info(f"Lecture réussie du fichier : {file_path}")
            if get_timer():
                elapsed_time = timeit.default_timer() - timer_start
                self.logger.info(f"Temps d'exécution de sqlQuery: {elapsed_time:.4f} secondes")
        except Exception as e:
            self.logger.error(f"Erreur de lecture du fichier SMB : {e}")
            raise

    def _read_local_file(self, file_path, mode, encoding):
        """
        Lit un fichier en local en batchs
        """
        timer_start = timeit.default_timer()
        try:
            with open(file_path, mode=mode, encoding=encoding) as file:
                while chunk := file.read(self.batch_size):
                    yield chunk
                self.logger.info(f"Lecture réussie du fichier: {file_path}")
            if get_timer():
                elapsed_time = timeit.default_timer() - timer_start
                self.logger.info(f"Temps d'exécution de sqlQuery: {elapsed_time:.4f} secondes")
        except Exception as e:
            self.logger.error(f"Erreur de lecture du fichier en local: {e}")
            raise

    @timer
    def write_file(self, file_path: str, content, mode: str='wb', clean: bool=False):
        """
        Écrit du contenu dans un fichier local ou SMB en batchs
        """
        if clean:
            self.logger.info(f"Nettoyage du fichier {file_path}.")
        if self.smb_config and file_path.startswith("//"):
            return self._write_smb_file(file_path, content, mode, clean)
        return self._write_local_file(file_path, content, mode, clean)

    def _write_smb_file(self, file_path, content, mode, clean: bool=False):
        """
        Écrit du contenu dans un fichier via partage SMB en batchs
        """
        try:
            encoding = "utf-8" if "b" not in mode else None
            with smbclient.open_file(file_path, mode=mode, encoding=encoding) as file:
                for chunk in content:
                    if clean:
                        chunk = self.clean_text(chunk)
                    for i in range(0, len(chunk), self.batch_size):
                        file.write(chunk[i:i + self.batch_size])
                self.logger.info(f"Succès : contenu écrit dans '{file_path}' (SMB).")
        except Exception as e:
            self.logger.error(f"Erreur lors de l'écriture du fichier '{file_path}' via SMB: {e}")
            raise

    def _write_local_file(self, file_path, content, mode, clean: bool=False):
        """
        Écrit du contenu dans un fichier local en batchs
        """
        try:
            encoding = "utf-8-sig" if "b" not in mode else None
            with open(file_path, mode=mode, encoding=encoding) as file:
                for chunk in content:
                    if clean:
                        chunk = self.clean_text(chunk)
                    for i in range(0, len(chunk), self.batch_size):
                        file.write(chunk[i:i + self.batch_size])
                self.logger.info(f"Succès: Contenu écrit dans '{file_path}' (local).")
        except Exception as e:
            self.logger.error(f"Erreur lors de l'écriture du fichier '{file_path}' en local: {e}")
            raise

    def list_dir(self, dir_path: str):
        """
        Liste le contenu d'un dossier en local ou via SMB
        """
        try:
            if self.smb_config and dir_path.startswith("//"):
                return smbclient.listdir(dir_path)
            return os.listdir(dir_path)
        except Exception as e:
            self.logger.error(f"Erreur lors de la lecture du dossier {dir_path}: {e}")
            return []

    def file_exists(self, file_path: str):
        """
        Vérifie si le fichier existe (local ou SMB)
        """
        if self.smb_config and file_path.startswith("//"):
            return self._smb_file_exists(file_path)
        return self._local_file_exists(file_path)

    def _smb_file_exists(self, file_path):
        """
        Vérifie si un fichier SMB existe
        """
        try:
            exists = smbclient.path.exists(file_path) and smbclient.path.isfile(file_path)
            if exists:
                self.logger.info(f"Le fichier existe: {file_path}")
            else:
                self.logger.warning(f"Le fichier n'existe pas: {file_path}")
            return exists
        except Exception as e:
            self.logger.error(f"Erreur lors de la vérification d'existence du fichier SMB: {e}")

    def _local_file_exists(self, file_path):
        """
        Vérifie si un fichier en local existe
        """
        exists = os.path.exists(file_path) and os.path.isfile(file_path)
        if exists:
            self.logger.info(f"Le fichier existe: {file_path}")
        else:
            self.logger.warning(f"Le fichier n'existe pas: {file_path}")
        return exists

    def directory_exists(self, dir_path: str):
        """
        Vérifie si le fichier existe (local ou SMB)
        """
        if self.smb_config and dir_path.startswith("//"):
            return self._smb_directory_exists(dir_path)
        return self._local_directory_exists(dir_path)

    def _smb_directory_exists(self, dir_path: str):
        try:
            exists = smbclient.path.exists(dir_path) and smbclient.path.isdir(dir_path)
            if exists:
                self.logger.info(f"Le dossier existe: {dir_path}")
            else:
                self.logger.warning(f"Le dossier n'existe pas: {dir_path}")
            return exists
        except Exception as e:
            self.logger.error(f"Erreur lors de la vérification du dossier SMB: {e}")

    def _local_directory_exists(self, dir_path):
        exists = os.path.exists(dir_path) and os.path.isdir(dir_path)
        if exists:
            self.logger.info(f"Le dossier existe: {dir_path}")
        else:
            self.logger.warning(f"Le dossier n'existe pas: {dir_path}")
        return exists

    def remove_file(self, file_path: str):
        """
        Supprime un fichier (local ou SMB)
        """
        try:
            if self.smb_config and file_path.startswith("//"):
                smbclient.remove(file_path)
            else:
                os.remove(file_path)
            self.logger.info(f"Fichier supprimé avec succès.")
        except Exception as e:
            self.logger.error(f"Erreur lors de la suppression du fichier '{file_path}': {e}")
            raise

    def create_empty_file(self, file_path: str, mode: str='wb'):
        """
        Crée un fichier vide (local ou SMB)
        """
        try:
            if self.smb_config and file_path.startswith("//"):
                with smbclient.open_file(file_path, mode=mode) as file:
                    pass
                self.logger.info(f"Fichier vide créé sur SMB: {file_path}")
            else:
                with open(file_path, mode=mode) as file:
                    pass
                self.logger.info(f"Fichier vide créé localement: {file_path}")
        except Exception as e:
            self.logger.error(f"Erreur lors de la création du fichier vide '{file_path}': {e}")
            raise

    def clean_text(self, text):
        if isinstance(text, bytes):
            encoding = chardet.detect(text)["encoding"]
            text = text.decode(encoding or "utf-8", errors='ignore')
        text = unicodedata.normalize('NFKD', text)
        text = ''.join([c for c in text if not unicodedata.combining(c)])
        text = ''.join([f'&#{ord(c)};' if ord(c) > 127 else c for c in text])
        return text.encode("utf-8")


import os
import mysql.connector
import timeit
import polars as pl

from .timer import timer, get_timer
from .dataFactory import data_factory

class dbMysql(data_factory):
    def __init__(self, dictionnary: dict, logger, batch_size=10_000):
        """
        instancie la classe dbMysql, qui hérite de dataFactory
        :param dictionnary: Un dictionnaire contenant tous les paramètres nécéssaires pour lancer une connexion MySQL
        :param logger: Un logger ads qui va gérer les logs des actions de la classe
        :param batch_size: La taille des batchs en lecture et écriture
        """
        self.connection = None
        self.logger = logger
        self.__database = dictionnary.get("database")
        self.__user = dictionnary.get("user")
        self.__password = dictionnary.get("password")
        self.__port = dictionnary.get("port")
        self.__host = dictionnary.get("host")
        self.batch_size = batch_size

    def connect(self, additionnal_parameters=None):
        """
        Lance la connexion avec les identifiants passés à l'initialisation de la classe
        Toutes les méthodes de la classe nécéssitent une connexion active
        :return: La connexion
        """
        if self.logger is not None: self.logger.info("Tentative de connexion avec la base de données.")
        try:
            self.connection = mysql.connector.connect(**{
                'user': self.__user,
                'password': self.__password,
                'host': self.__host,
                'port': self.__port,
                'database': self.__database,
            })
            if additionnal_parameters:
                cursor = self.connection.cursor()
                for param, value in additionnal_parameters.items():
                    cursor.execute(f"SET SESSION {param} = {value}")
                self.connection.commit()
            if self.logger is not None: self.logger.info("Connexion établie avec la base de données.")
            return self.connection
        except Exception as e:
            if self.logger is not None: self.logger.error(f"Échec de la connexion à la base de données: {e}")
            raise

    def sqlQuery(self, query: str, return_columns: bool = False):
        """
        lit la base de données avec la requête query
        :param return_columns: booléen qui indique si l'on veut récupérer les colonnes de la tables
        :param query: la requête
        :return: les données lues avec yield
        """
        self.logger.debug(f"Exécution de la requête de lecture: {query}")
        try:
            timer_start = timeit.default_timer()
            cpt_rows = 0
            with self.connection.cursor() as cursor:
                cursor.execute(query)
                if return_columns:
                    cols = [desc[0] for desc in cursor.description]
                    yield cols
                self.logger.info("Requête exécutée avec succès, début de la lecture des résultats.")
                while True:
                    rows = cursor.fetchmany(self.batch_size)
                    if not rows:
                        break
                    yield rows
                    cpt_rows += len(rows)
                    self.logger.info(f"{cpt_rows} ligne(s) lue(s).")
            if get_timer():
                elapsed_time = timeit.default_timer() - timer_start
                self.logger.info(f"Temps d'exécution de sqlQuery: {elapsed_time:.4f} secondes.")
        except Exception as e:
            self.logger.error(f"Échec de la lecture des données: {e}")
            raise

    @timer
    def sqlExec(self, query):
        """
        Exécute une requête sur la base de données, un create ou delete table par exemple
        :param query: La requête en question
        """
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query)
                self.connection.commit()
                self.logger.info(f"Requête exécutée avec succès.")
        except Exception as e:
            self.logger.error(f"Échec de l'exécution de la requête: {e}")
            raise

    @timer
    def sqlScalaire(self, query):
        """
        Exécute une requête et retourne le premier résultat
        :param query: La requête en question
        :return: Le résultat de cette requête
        """
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query)
                result = cursor.fetchone()
                self.logger.info(f"Requête scalaire exécutée avec succès.")
                return result[0] if result else None
        except Exception as e:
            self.logger.error(f"Échec de l'exécution de la requête: {e}")
            raise

    @timer
    def insert(self, schema: str, table: str, cols: [], row: []):
        """
        Insère une ligne de données dans la base de données
        :param table: Nom de la table dans laquelle insérer
        :param cols: Liste des colonnes dans lesquelles insérer
        :param row: Liste des valeurs à insérer
        :return: Le résultat de l'opération
        """
        placeholders = ', '.join(["%s"] * len(cols))
        table = f"{schema}.{table}" if schema else table
        query = f"INSERT INTO {table} ({', '.join(cols)}) VALUES ({placeholders})"
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query, row)
                self.connection.commit()
                self.logger.info(f"{len(row)} valeurs insérées avec succès dans la table {table}")
                return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Échec de l'insertion des données: {e}")
            return "ERROR", str(e), row

    @timer
    def insertMany(self, schema: str, table: str, cols: [], rows: [[]]):
        """
        Insère des données par batch dans une table avec gestion des erreurs

        :param schema: Nom du schéma dans lequel se trouve la table
        :param table: Nom de la table dans laquelle insérer
        :param cols: Liste des colonnes dans lesquelles insérer
        :param rows: Liste des lignes à insérer
        :return: Le résultat de l'opération, l'erreur et la le batch concerné en cas d'erreur
        """
        failed_batches = []
        errors = []
        total_inserted = 0
        placeholders = ', '.join(["%s"] * len(cols))
        table = f"{schema}.{table}" if schema else table
        query = f"INSERT INTO {table} ({', '.join(cols)}) VALUES ({placeholders})"
        try:
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, len(rows), self.batch_size), start=1):
                    batch = rows[start: start + self.batch_size]
                    try:
                        cursor.executemany(query, batch)
                        self.connection.commit()
                        total_inserted += len(batch)
                        self.logger.info(
                            f"Batch {batch_index}: {len(batch)} ligne(s) insérée(s) avec succès dans la table {table}. "
                            f"Total inséré: {total_inserted}/{len(rows)} ligne(s)."
                        )
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'insertion du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
                return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'insertion des données: {e}")
            return "ERROR", str(e), rows

    def __get_df(self, rows, cols):
        if isinstance(rows, list):
            df = pl.DataFrame(rows, schema=cols, orient='row', infer_schema_length=len(rows))
        elif isinstance(rows, pl.DataFrame):
            df = rows
        else:
            raise ValueError("Les données doivent être une liste de tuples ou un DataFrame polars")
        return df

    @timer
    def insertBulk(self, schema: str, table: str, cols: [], rows):
        """
        Insère des données dans une table en bulk, en lots

        :param schema: Nom du schéma
        :param table: Nom de la table dans laquelle insérer
        :param cols: Liste des colonnes dans lesquelles insérer
        :param rows: Liste des lignes à insérer
        :return: Le résultat de l'opération, l'erreur et la le batch concerné en cas d'erreur
        """
        failed_batches = []
        errors = []
        total_inserted = 0
        table_full = f"{schema}.{table}" if schema else table
        temp_file = "temp_bulk_insert.csv"
        query = f"""
        LOAD DATA LOCAL INFILE '{temp_file}'
        INTO TABLE {table_full}
        FIELDS TERMINATED BY ','
        LINES TERMINATED BY '\n'
        ({', '.join(cols)})
        """
        try:
            df = self.__get_df(rows, cols)
            n_rows = df.shape[0]
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, n_rows, self.batch_size), start=1):
                    batch = df.slice(start, self.batch_size)
                    try:
                        batch.write_csv(temp_file, include_header=False)
                        cursor.execute(query)
                        self.connection.commit()
                        os.remove(temp_file)
                        total_inserted += batch.shape[0]
                        self.logger.info(f"Batch {batch_index}: {batch.shape[0]} ligne(s) insérée(s) avec succès dans la table {table_full}. "
                                         f"Total inséré: {total_inserted}/{n_rows} ligne(s).")
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'insertion du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
            return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'insertion des données: {e}")
            return "ERROR", str(e), rows
        finally:
            if os.path.exists(temp_file):
                os.remove(temp_file)

    @timer
    def upsert(self, schema: str, table: str, cols: [], row: [], conflict_cols: []):
        """
        Réalise une opération upsert sur la base
        :param schema: Le schéma de la table
        :param table: Le nom de la table
        :param cols: Liste des colonnes dans lequelles insérer
        :param row: Liste des valeurs à insérer
        :param conflict_cols: Colonnes à utiliser pour détecter les conflits
        :return: Le résultat de l'opération, l'erreur et la ligne en cas d'erreur
        """
        table_full = f"{schema}.{table}" if schema else table
        columns = ', '.join([f"'{col}'" for col in cols])
        placeholders = ', '.join(['%s'] * len(cols))
        update_clause = ', '.join([f"'{col}' = VALUES('{col}')" for col in cols if col not in conflict_cols])
        query = f"""
        INSERT INTO {table_full} ({columns})
        VALUES ({placeholders})
        ON DUPLICATE KEY UPDATE {update_clause};
        """
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query, row)
                self.connection.commit()
                self.logger.info(f"{len(row)} valeurs insérée(s) ou mise(s) à jour dans la table {table_full}.")
                return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur lors de l'upsert dans la table {table}.")
            return "ERROR", str(e), row

    @timer
    def upsertMany(self, schema: str, table: str, cols: [], rows: [[]], conflict_cols: []):
        """
        Réalise une opération upsert par batch sur la base
        :param schema: Le schéma de la table
        :param table: Le nom de la table
        :param cols: Liste des colonnes dans lequelles insérer
        :param rows: Liste des lignes à insérer
        :param conflict_cols: Colonnes à utiliser pour détecter les conflits
        :return: Le résultat de l'opération, l'erreur et le batch en cas d'erreur
        """
        failed_batches = []
        errors = []
        total_inserted = 0
        table_full = f"{schema}.{table}" if schema else table
        columns = ', '.join([f"'{col}'" for col in cols])
        placeholders = ', '.join(['%s'] * len(cols))
        update_clause = ', '.join([f"'{col}' = VALUES('{col}')" for col in cols if col not in conflict_cols])
        query = f"""
        INSERT INTO {table_full} ({columns})
        VALUES ({placeholders})
        ON DUPLICATE KEY UPDATE {update_clause};
        """
        try:
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, len(rows), self.batch_size), start=1):
                    batch = rows[start: start + self.batch_size]
                    try:
                        cursor.executemany(query, batch)
                        self.connection.commit()
                        total_inserted += len(batch)
                        self.logger.info(
                            f"Batch {batch_index}: {len(batch)} ligne(s) insérée(s) ou mise(s) à jour dans la table {table}. "
                            f"Total inséré ou mis à jour: {total_inserted}/{len(rows)} ligne(s)."
                        )
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'upsert du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
            return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'upsert des données: {e}")
            return "ERROR", str(e), rows

    @timer
    def upsertBulk(self, schema: str, table: str, cols: [], rows, conflict_cols: []):
        """
        Réalise une opération upsert par batch sur la base avec LOAD DATA INFILE et une table temporaire.

        :param schema: Le schéma de la table
        :param table: Le nom de la table
        :param cols: Liste des colonnes dans lesquelles insérer
        :param rows: Liste des lignes à insérer
        :param conflict_cols: Colonnes à utiliser pour détecter les conflits
        :return: Le résultat de l'opération, les erreurs et les lots échoués
        """
        temp_table = f"{table}_temp"
        table_full, temp_table_full = (f"{schema}.{table}", f"{schema}.{temp_table}") if schema else (table, temp_table)
        columns = ', '.join(cols)
        update_clause = ', '.join([f"{col} = VALUES({col})" for col in cols if col not in conflict_cols])
        query = f"""
        INSERT INTO {table_full} ({columns})
        SELECT {columns} FROM {temp_table_full}
        ON DUPLICATE KEY UPDATE {update_clause};
        """
        try:
            self.sqlExec(f"CREATE TEMPORARY TABLE {temp_table_full} LIKE {table_full}")
            self.logger.info(f"Table temporaire {temp_table_full} créée.")
            result, errors, failed_batches = self.insertBulk(schema, temp_table, cols, rows)
            if result == "ERROR":
                return result, errors, failed_batches
            self.sqlExec(query)
            self.logger.info(f"{len(rows)} lignes insérée(s) ou mise(s) à jour dans la table {table}.")
            return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'UPSERT : {e}")
            return "ERROR", str(e), rows
        finally:
            self.sqlExec(f"DROP TEMPORARY TABLE IF EXISTS {temp_table_full}")
            self.logger.info(f"Table temporaire {temp_table_full} supprimée.")
